#! /usr/bin/python

# pacman.pyw
# By David Reilly

# Modified by Andy Sommerville, 8 October 2007:
# - Changed hard-coded DOS paths to os.path calls
# - Added constant SCRIPT_PATH (so you don't need to have pacman.pyw and res in your cwd, as long
# -   as those two are in the same directory)
# - Changed text-file reading to accomodate any known EOLn method (\n, \r, or \r\n)
# - I (happily) don't have a Windows box to test this. Blocks marked "WIN???"
# -   should be examined if this doesn't run in Windows
# - Added joystick support (configure by changing JS_* constants)
# - Added a high-score list. Depends on wx for querying the user's name

import sys, os, random
from direct.directbase.DirectStart import *
from direct.fsm.FSM import FSM
from direct.showbase.PythonUtil import fitSrcAngle2Dest
from direct.gui.DirectGui import *
from direct.stdpy.file import open
from panda3d.core import *

creditsList = """
\1H\1Coding:\2
David Reilly
David Rose
Andy Sommerville

\1H\1Models/costumes:\2
Belle Benson
Nara Dunkerson
Kent Elofson
Nicole Roberts
Cordelia Willis

\1H\1Photography:\2
Eric Pearce

\1H\1With apologies to Namco\2
"""

if base.appRunner:
    # Packed p3d file
    SCRIPT_PATH = base.appRunner.multifileRoot
else:
    # Interactive Python file
    SCRIPT_PATH = Filename.fromOsSpecific(sys.path[0])
    getModelPath().prependDirectory(SCRIPT_PATH)

# NO_GIF_TILES -- tile numbers which do not correspond to a GIF file
# currently only "23" for the high-score list
NO_GIF_TILES=[23]

NO_WX=0 # if set, the high-score code will not attempt to ask the user his name
USER_NAME="User" # USER_NAME=os.getlogin() # the default user name if wx fails to load or NO_WX

# Joystick defaults - maybe add a Preferences dialog in the future?
JS_DEVNUM=0 # device 0 (pygame joysticks always start at 0). if JS_DEVNUM is not a valid device, will use 0
JS_XAXIS=0 # axis 0 for left/right (default for most joysticks)
JS_YAXIS=1 # axis 1 for up/down (default for most joysticks)
JS_STARTBUTTON=0 # button number to start the game. this is a matter of personal preference, and will vary from device to device

font = loader.loadFont('res/prstart.ttf')
pacFont = loader.loadFont('res/PacFont Good.ttf', pixelsPerUnit = 50)

tpMgr = TextPropertiesManager.getGlobalPtr()
tph = TextProperties()
tph.setTextColor(0.5, 0.5, 1.0, 1.0)
tpMgr.setProperties('H', tph)

# Get the Batman logo to use in the title.
model = loader.loadModel('res/text/text')
logo = model.find('**/life')
tpMgr.setGraphic('LOGO', logo)

snd_pellet = {}
snd_pellet[0] = loader.loadSfx("res/sounds/pellet1.wav")
snd_pellet[1] = loader.loadSfx("res/sounds/pellet2.wav")
snd_powerpellet = loader.loadSfx("res/sounds/powerpellet.wav")
snd_eatgh = loader.loadSfx("res/sounds/eatgh2.wav")
snd_fruitbounce = loader.loadSfx("res/sounds/fruitbounce.wav")
snd_eatfruit = loader.loadSfx("res/sounds/eatfruit.wav")
snd_extralife = loader.loadSfx("res/sounds/extralife.wav")

ghostcolor = {}
ghostcolor[0] = (255, 0, 0, 255)
ghostcolor[1] = (255, 128, 255, 255)
ghostcolor[2] = (128, 255, 255, 255)
ghostcolor[3] = (255, 128, 0, 255)
ghostcolor[4] = (50, 50, 255, 255) # blue, vulnerable ghost
ghostcolor[5] = (255, 255, 255, 255) # white, flashing ghost

ghostmodel = {}
ghostmodel[0] = 'catwoman'
ghostmodel[1] = 'talia'
ghostmodel[2] = 'poison_ivy'
ghostmodel[3] = 'harley_quinn'

#      ___________________
# ___/  class definitions  \_______________________________________________

class game (FSM):

    def defaulthiscorelist(self):
        return [ (100000,"David") , (80000,"Andy") , (60000,"Count Pacula") , (40000,"Cleopacra") , (20000,"Brett Favre") , (10000,"Sergei Pachmaninoff") ]

    def getHiScores(self):
        """If res/hiscore.txt exists, read it. If not, return the default high scores.
           Output is [ (score,name) , (score,name) , .. ]. Always 6 entries."""
        try:
          f=open("pacman_hiscore.txt")
          hs=[]
          for line in f:
            while len(line)>0 and (line[0]=="\n" or line[0]=="\r"): line=line[1:]
            while len(line)>0 and (line[-1]=="\n" or line[-1]=="\r"): line=line[:-1]
            score=int(line.split(" ")[0])
            name=line.partition(" ")[2]
            if score>99999999: score=99999999
            if len(name)>22: name=name[:22]
            hs.append((score,name))
          f.close()
          if len(hs)>6: hs=hs[:6]
          while len(hs)<6: hs.append((0,""))
          return hs
        except IOError:
          return self.defaulthiscorelist()
              
    def writeHiScores(self, hs):
        """Given a new list, write it to the default file."""
        f=open("pacman_hiscore.txt","w")
        for line in hs:
            f.write(str(line[0])+" "+line[1]+"\n")
        f.close()

    def isHighScore(self, newscore):
        """ Returns true if the score qualifies for the high-score
        list, false otherwise. """

        for line in self.hiScores:
            if newscore >= line[0]:
                return True
        return False
              
    def recordHiScore(self, newscore, name):
        """Add newscore to the high score list, if appropriate."""
        hs = self.getHiScores()
        for line in hs:
            if newscore >= line[0]:
                hs.insert(hs.index(line), (newscore, name))
                hs.pop(-1)
                self.writeHiScores(hs)
                break
        self.hiScores = hs
        self.makeHiScoreList()

    def makeHiScoreList(self):
        "Read the High-Score file and apply it to self.highScoreList."

        lines = []
        for line in self.hiScores:
            lines.append(line[1].upper().ljust(22)+str(line[0]).rjust(9))

        self.highScoreList.setText('\n'.join(lines))

    def __init__ (self):
        FSM.__init__(self, 'game')
        self.levelNum = 0
        self.score = 0
        self.lives = 3

        # game "mode" variable
        # 1 = Normal
        # 2 = HitGhost
        # 3 = GameOver
        # 4 = WaitToStart
        # 5 = WaitAfterEatingGhost
        # 6 = WaitAfterFinishing
        # 7 = FlashingMaze
        # 8 = BlankScreen
        # 9 = PromptForName
        self.mode = 0
        self.ghostTimer = 0
        self.ghostValue = 0
        self.fruitTimer = 0
        self.fruitScoreTimer = 0
        self.fruitScorePos = (0, 0)

        scoreSize = 0.05
        self.scoreLabel = OnscreenText(parent = base.a2dTopRight, font = font, align = TextNode.ARight, fg = (1, 1, 1, 1), scale = scoreSize, pos = (-0.1, -scoreSize))

        # camera variables
        self.screenPixelPos = (0, 0) # absolute x,y position of the screen from the upper-left corner of the level
        self.screenNearestTilePos = (0, 0) # nearest-tile position of the screen from the UL corner
        self.screenPixelOffset = (0, 0) # offset in pixels of the screen from its nearest-tile position
        
        self.screenTileSize = (23, 21)
        self.screenSize = (self.screenTileSize[1] * 16, self.screenTileSize[0] * 16)

        model = loader.loadModel('res/text/text')
        logo = model.find('**/life')

        lifeSize = 0.15
        lifeSwitch = NodePath(SwitchNode('lifeSwitch'))
        for i in range(8):
            child = lifeSwitch.attachNewNode('life_%s' % (i))
            for j in range(i):
                np = logo.copyTo(child)
                np.setPos(j, 0, 0)
        lifeSwitch.reparentTo(base.a2dTopLeft)
        lifeSwitch.setScale(lifeSize)
        lifeSwitch.setPos(0.1, 0, -0.05)
        self.lifeSwitch = lifeSwitch
        self.lifeSwitchNode = lifeSwitch.node()

        self.board = render.attachNewNode('board')
        self.setupTitlePage()
        self.setupCreditsPage()
        self.setupNameEntryPage()

        base.accept('escape', self.escapeKey)

        self.request('GameOver')

    def escapeKey(self):
        if self.state != 'GameOver':
            self.request('GameOver')
            return

        # In the GameOver state, escape exits, but only if we're not
        # running in a browser.
        if not base.appRunner or not base.appRunner.dom:
            sys.exit(0)

    def showGame(self):
        self.lifeSwitch.show()
        self.scoreLabel.show()
        self.board.show()

    def hideGame(self):
        self.lifeSwitch.hide()
        self.scoreLabel.hide()
        self.board.hide()

    def enterNormal(self):
        self.lastMoveFrame = int(globalClock.getFrameTime() * 60.0)

    def enterHitGhost(self):
        taskMgr.doMethodLater(1.5, self.continueHitGhost, 'continueHitGhost')

    def exitHitGhost(self):
        taskMgr.remove('continueHitGhost')

    def continueHitGhost(self, task):
        thisLevel.Restart()

        thisGame.lives -= 1
        if thisGame.lives == -1:
            if thisGame.isHighScore(thisGame.score):
                thisGame.request('PromptForName')
            else:
                thisGame.request('GameOver')
        else:
            thisGame.request('WaitToStart')

    def enterGameOver(self):
        self.hideGame()
        self.titlePage.show()
        base.accept('f1', self.request, extraArgs = ['Credits'])
        base.accept('enter', self.StartNewGame)

    def exitGameOver(self):
        self.titlePage.hide()
        self.showGame()
        base.ignore('f1')
        base.ignore('enter')

    def enterCredits(self):
        self.hideGame()
        self.creditsPage.show()
        base.accept('enter', self.request, extraArgs = ['GameOver'])

    def exitCredits(self):
        self.creditsPage.hide()
        self.showGame()
        base.ignore('enter')

    def enterWaitToStart(self):
        taskMgr.doMethodLater(1.5, self.continueWaitToStart, 'continueWaitToStart')

    def exitWaitToStart(self):
        taskMgr.remove('continueWaitToStart')

    def continueWaitToStart(self, task):
        self.request('Normal')
        player.velX = player.speed

    def enterWaitAfterEatingGhost(self):
        taskMgr.doMethodLater(0.5, self.continueWaitAfterEatingGhost, 'continueWaitAfterEatingGhost')

    def exitWaitAfterEatingGhost(self):
        taskMgr.remove('continueWaitAfterEatingGhost')

    def continueWaitAfterEatingGhost(self, task):
        self.request('Normal')

    def enterWaitAfterFinishing(self):
        taskMgr.doMethodLater(1.0, self.continueAfterFinishing, 'continueWaitAfterFinishing')

    def exitWaitAfterFinishing(self):
        taskMgr.remove('continueWaitAfterFinishing')

    def continueAfterFinishing(self, task):
        self.request('FlashingMaze')

    def enterBlankScreen(self):
        self.board.hide()
        taskMgr.doMethodLater(0.67, self.SetNextLevel, 'continueBlankScreen', extraArgs = [])

    def exitBlankScreen(self):
        self.board.show()
        taskMgr.remove('continueBlankScreen')

    def enterFlashingMaze(self):
        taskMgr.add(self.doFlashingMaze, 'doFlashingMaze')

    def exitFlashingMaze(self):
        taskMgr.remove('doFlashingMaze')

    def doFlashingMaze(self, task):
        elapsed = task.getElapsedTime()
        frame = int(elapsed * 4.0)
        if frame % 2 == 0:
            # member of white set
            thisLevel.mapCard.setTexture(thisLevel.mapTexture1, 1)
        else:
            # member of normal set
            thisLevel.mapCard.setTexture(thisLevel.mapTexture0, 1)

        if elapsed > 2.5:
            self.request('BlankScreen')

        return task.cont

    def enterPromptForName(self):
        self.hideGame()
        self.nameEntryPage.show()
        self.nameEntry['focus'] = True

    def exitPromptForName(self):
        self.nameEntryPage.hide()
        self.showGame()
        
    def setupTitlePage(self):
        self.titlePage = aspect2d.attachNewNode('titlePage')
        self.titlePage.hide()
        title = OnscreenText(text = 'pac\5LOGO\5bat', font = pacFont, parent = self.titlePage, fg = (1, 1, 0, 1), scale = 0.25, pos = (0, 0.6))

        gameOver = OnscreenText(text = 'GAME OVER', font = font, parent = self.titlePage, fg = (1, 0, 0, 1), scale = 0.05, pos = (0, 0.4))
        toStart = OnscreenText(text = 'PRESS ENTER TO START', font = font, parent = self.titlePage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, 0.3))

        self.highScoreHeader = OnscreenText(text = 'HIGH SCORES', font = font, parent = self.titlePage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, 0.1))
        
        self.highScoreList = OnscreenText(font = font, parent = self.titlePage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, 0))
        self.hiScores = self.getHiScores()
        self.makeHiScoreList()

        toCredits = OnscreenText(text = 'PRESS F1 FOR CREDITS', font = font, parent = self.titlePage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, -0.8))

    def setupCreditsPage(self):
        self.creditsPage = aspect2d.attachNewNode('creditsPage')
        self.creditsPage.hide()
        title = OnscreenText(text = 'credits', font = pacFont, parent = self.creditsPage, fg = (1, 1, 0, 1), scale = 0.25, pos = (0, 0.6))

        text = creditsList.strip().upper()
        credits = OnscreenText(text = text, font = font, parent = self.creditsPage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, 0.4))

        toReturn = OnscreenText(text = 'PRESS ENTER TO RETURN', font = font, parent = self.creditsPage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, -0.8))

    def setupNameEntryPage(self):
        self.nameEntryPage = aspect2d.attachNewNode('nameEntryPage')
        self.nameEntryPage.hide()
        title = OnscreenText(text = 'awesome!', font = pacFont, parent = self.nameEntryPage, fg = (1, 1, 0, 1), scale = 0.25, pos = (0, 0.6))

        prompt = OnscreenText(text = "YOU MADE THE HIGH SCORE LIST!\nPLEASE ENTER YOUR NAME.", font = font, parent = self.nameEntryPage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, 0.4))

        self.nameEntry = DirectEntry(
            width = 22, entryFont = font, focus = True, scale = 0.05,
            pos = (-0.5, 0, 0), command = self.enterName,
            parent = self.nameEntryPage, suppressKeys = True)
        self.nameEntry.guiItem.getTextDef(0).setSmallCaps(True)
        self.nameEntry.set('')

        toReturn = OnscreenText(text = 'PRESS ENTER WHEN DONE', font = font, parent = self.nameEntryPage, fg = (1, 1, 1, 1), scale = 0.05, pos = (0, -0.8))

    def enterName(self, name):
        name = name.strip()
        if not name:
            # Go back and wait for more text.
            self.nameEntry['focus'] = True
            return
        
        self.recordHiScore(self.score, name)
        self.request('GameOver')
        
    def StartNewGame (self):
        self.levelNum = 1
        self.score = 0
        self.lives = 3
        
        self.request('WaitToStart')
        thisLevel.LoadLevel( thisGame.GetLevelNum() )
            
    def AddToScore (self, amount, drawAt = None):
        
        extraLifeSet = [25000, 50000, 100000, 150000]
        
        for specialScore in extraLifeSet:
            if self.score < specialScore and self.score + amount >= specialScore:
                snd_extralife.play()
                thisGame.lives += 1
        
        self.score += amount

        if drawAt:
            x, y = drawAt
            label = OnscreenText(text = str(amount), parent = render, font = font, align = TextNode.ACenter, fg = (1, 1, 1, 1), scale = 0.2)
            label.setBillboardPointEye()
            label.setDepthTest(False)
            label.setBin('fixed', 0)
            NodePath.setPos(label, thisLevel.pixelBoard, x + 8, y + 8, 0)
            taskMgr.doMethodLater(2, self.removeLabel, 'removeLabel', extraArgs = [label])

    def removeLabel(self, label):
        label.removeNode()
    
    def DrawScore (self):
        self.scoreLabel.setText(str(self.score))
        self.lifeSwitchNode.setVisibleChild(max(self.lives, 0))

    def GetScreenPos (self):
        return self.screenPixelPos
        
    def GetLevelNum (self):
        return self.levelNum
    
    def SetNextLevel (self):
        self.levelNum += 1
        
        self.request('WaitToStart')
        thisLevel.LoadLevel( thisGame.GetLevelNum() )
        
        player.velX = 0
        player.velY = 0
        
class node ():
    
    def __init__ (self):
        self.g = -1 # movement cost to move from previous node to this one (usually +10)
        self.h = -1 # estimated movement cost to move from this node to the ending node (remaining horizontal and vertical steps * 10)
        self.f = -1 # total movement cost of this node (= g + h)
        # parent node - used to trace path back to the starting node at the end
        self.parent = (-1, -1)
        # node type - 0 for empty space, 1 for wall (optionally, 2 for starting node and 3 for end)
        self.type = -1
        
class path_finder ():
    
    def __init__ (self):
        # map is a 1-DIMENSIONAL array.
        # use the Unfold( (row, col) ) function to convert a 2D coordinate pair
        # into a 1D index to use with this array.
        self.map = {}
        self.size = (-1, -1) # rows by columns
        
        self.pathChainRev = ""
        self.pathChain = ""
                
        # starting and ending nodes
        self.start = (-1, -1)
        self.end = (-1, -1)
        
        # current node (used by algorithm)
        self.current = (-1, -1)
        
        # open and closed lists of nodes to consider (used by algorithm)
        self.openList = []
        self.closedList = []
        
        # used in algorithm (adjacent neighbors path finder is allowed to consider)
        self.neighborSet = [ (0, -1), (0, 1), (-1, 0), (1, 0) ]
        
    def ResizeMap (self, (numRows, numCols)):
        self.map = {}
        self.size = (numRows, numCols)

        # initialize path_finder map to a 2D array of empty nodes
        for row in range(0, self.size[0], 1):
            for col in range(0, self.size[1], 1):
                self.Set( (row, col), node() )
                self.SetType( (row, col), 0 )
        
    def CleanUpTemp (self):
        
        # this resets variables needed for a search (but preserves the same map / maze)
    
        self.pathChainRev = ""
        self.pathChain = ""
        self.current = (-1, -1)
        self.openList = []
        self.closedList = []
        
    def FindPath (self, startPos, endPos ):
        
        self.CleanUpTemp()
        
        # (row, col) tuples
        self.start = startPos
        self.end = endPos
        
        # add start node to open list
        self.AddToOpenList( self.start )
        self.SetG ( self.start, 0 )
        self.SetH ( self.start, 0 )
        self.SetF ( self.start, 0 )
        
        doContinue = True
        
        while (doContinue == True):
        
            thisLowestFNode = self.GetLowestFNode()

            if not thisLowestFNode == self.end and not thisLowestFNode == False:
                self.current = thisLowestFNode
                self.RemoveFromOpenList( self.current )
                self.AddToClosedList( self.current )
                
                for offset in self.neighborSet:
                    thisNeighbor = (self.current[0] + offset[0], self.current[1] + offset[1])
                    
                    if not thisNeighbor[0] < 0 and not thisNeighbor[1] < 0 and not thisNeighbor[0] > self.size[0] - 1 and not thisNeighbor[1] > self.size[1] - 1 and not self.GetType( thisNeighbor ) == 1:
                        cost = self.GetG( self.current ) + 10
                        
                        if self.IsInOpenList( thisNeighbor ) and cost < self.GetG( thisNeighbor ):
                            self.RemoveFromOpenList( thisNeighbor )
                            
                        #if self.IsInClosedList( thisNeighbor ) and cost < self.GetG( thisNeighbor ):
                        #   self.RemoveFromClosedList( thisNeighbor )
                            
                        if not self.IsInOpenList( thisNeighbor ) and not self.IsInClosedList( thisNeighbor ):
                            self.AddToOpenList( thisNeighbor )
                            self.SetG( thisNeighbor, cost )
                            self.CalcH( thisNeighbor )
                            self.CalcF( thisNeighbor )
                            self.SetParent( thisNeighbor, self.current )
            else:
                doContinue = False
                        
        if thisLowestFNode == False:
            return False
                        
        # reconstruct path
        self.current = self.end
        while not self.current == self.start:
            # build a string representation of the path using R, L, D, U
            if self.current[1] > self.GetParent(self.current)[1]:
                self.pathChainRev += 'R' 
            elif self.current[1] < self.GetParent(self.current)[1]:
                self.pathChainRev += 'L'
            elif self.current[0] > self.GetParent(self.current)[0]:
                self.pathChainRev += 'D'
            elif self.current[0] < self.GetParent(self.current)[0]:
                self.pathChainRev += 'U'
            self.current = self.GetParent(self.current)
            self.SetType( self.current, 4)
            
        # because pathChainRev was constructed in reverse order, it needs to be reversed!
        for i in range(len(self.pathChainRev) - 1, -1, -1):
            self.pathChain += self.pathChainRev[i]
        
        # set start and ending positions for future reference
        self.SetType( self.start, 2)
        self.SetType( self.end, 3)
        
        return self.pathChain

    def Unfold (self, (row, col)):
        # this function converts a 2D array coordinate pair (row, col)
        # to a 1D-array index, for the object's 1D map array.
        return (row * self.size[1]) + col
    
    def Set (self, (row, col), newNode):
        # sets the value of a particular map cell (usually refers to a node object)
        self.map[ self.Unfold((row, col)) ] = newNode
        
    def GetType (self, (row, col)):
        return self.map[ self.Unfold((row, col)) ].type
        
    def SetType (self, (row, col), newValue):
        self.map[ self.Unfold((row, col)) ].type = newValue

    def GetF (self, (row, col)):
        return self.map[ self.Unfold((row, col)) ].f

    def GetG (self, (row, col)):
        return self.map[ self.Unfold((row, col)) ].g
    
    def GetH (self, (row, col)):
        return self.map[ self.Unfold((row, col)) ].h
        
    def SetG (self, (row, col), newValue ):
        self.map[ self.Unfold((row, col)) ].g = newValue

    def SetH (self, (row, col), newValue ):
        self.map[ self.Unfold((row, col)) ].h = newValue
        
    def SetF (self, (row, col), newValue ):
        self.map[ self.Unfold((row, col)) ].f = newValue
        
    def CalcH (self, (row, col)):
        self.map[ self.Unfold((row, col)) ].h = abs(row - self.end[0]) + abs(col - self.end[0])
        
    def CalcF (self, (row, col)):
        unfoldIndex = self.Unfold((row, col))
        self.map[unfoldIndex].f = self.map[unfoldIndex].g + self.map[unfoldIndex].h
    
    def AddToOpenList (self, (row, col) ):
        self.openList.append( (row, col) )
        
    def RemoveFromOpenList (self, (row, col) ):
        self.openList.remove( (row, col) )
        
    def IsInOpenList (self, (row, col) ):
        if self.openList.count( (row, col) ) > 0:
            return True
        else:
            return False
        
    def GetLowestFNode (self):
        lowestValue = 1000 # start arbitrarily high
        lowestPair = (-1, -1)
        
        for iOrderedPair in self.openList:
            if self.GetF( iOrderedPair ) < lowestValue:
                lowestValue = self.GetF( iOrderedPair )
                lowestPair = iOrderedPair
        
        if not lowestPair == (-1, -1):
            return lowestPair
        else:
            return False
        
    def AddToClosedList (self, (row, col) ):
        self.closedList.append( (row, col) )
        
    def IsInClosedList (self, (row, col) ):
        if self.closedList.count( (row, col) ) > 0:
            return True
        else:
            return False

    def SetParent (self, (row, col), (parentRow, parentCol) ):
        self.map[ self.Unfold((row, col)) ].parent = (parentRow, parentCol)

    def GetParent (self, (row, col) ):
        return self.map[ self.Unfold((row, col)) ].parent

class threeSixty:
    rotationSpeed = 400  # degrees per second
    
    def __init__(self, modelName):
        self.modelName = modelName
        model = loader.loadModel('res/360s/' + modelName)
        seq = model.find('**/+SequenceNode')
        frames = seq.getChildren()

        self.model = NodePath('model')
        self.switch = self.model.attachNewNode(SwitchNode(modelName))
        self.switch.setPos(0, 0, -0.2)
        self.switch.setBillboardAxis()
        self.switchNode = self.switch.node()
        frames.reparentTo(self.switch)
        self.numFrames = len(frames)

        self.targetHeading = 0
        self.interval = None

    def setTargetHeading(self, h):
        if h == self.targetHeading:
            # No change.
            return
        self.targetHeading = h

        if self.interval:
            self.interval.pause()

        heading = fitSrcAngle2Dest(self.model.getH(), self.targetHeading)
        delta = abs(self.targetHeading - heading)
        duration = delta / self.rotationSpeed
        
        # Create an interval to rotate us to the desired heading.
        self.interval = self.model.hprInterval(duration, (self.targetHeading, 0, 0),
                                               startHpr = (heading, 0, 0))
        self.interval.start()

    def update(self):
        # Selects the appropriate face to show the camera.
        h = self.model.getH(base.cam)
        index = int(self.numFrames * (h / -360.0 + 0.5) + 0.5) % self.numFrames
        assert index >= 0
        self.switchNode.setVisibleChild(index)

class ghost (FSM):
    def __init__ (self, ghostID):
        FSM.__init__(self, 'ghost_%s' % (ghostID))
        self.x = 0
        self.y = 0
        self.velX = 0
        self.velY = 0
        self.speed = 1
        
        self.nearestRow = 0
        self.nearestCol = 0
        
        self.id = ghostID
        
        self.homeX = 0
        self.homeY = 0

        if self.id in ghostmodel:
            self.model = threeSixty(ghostmodel[self.id])
            self.model.model.setScale(0.2)
        
        self.currentPath = ""
            
        self.animFrame = 1
        self.animDelay = 0

        self.startFlashTime = 0
        self.endVulnerableTime = 0

        # ghost "state" variable
        # 1 = normal
        # 2 = vulnerable
        # 3 = spectacles
        self.stateId = 1

    def enterOff(self):
        self.model.model.detachNode()

    def enterNormal(self):
        self.model.model.reparentTo(thisLevel.board)

    def enterVulnerable(self):
        self.model.model.reparentTo(thisLevel.board)
        self.model.model.setColor(0, 0, 1, 1)
        now = globalClock.getFrameTime()
        self.endVulnerableTime = now + 8.0
        self.startFlashTime = self.endVulnerableTime - 3.0

    def exitVulnerable(self):
        self.model.model.clearColor()
        self.ghostValue = 0

    def Draw (self):
        if thisGame.state == 'GameOver':
            self.request('Off')
            return

        if self.state == 'Vulnerable':
            now = globalClock.getFrameTime()
            if now > self.endVulnerableTime:
                self.request('Normal')
            elif now > self.startFlashTime:
                flash = int((now - self.startFlashTime) * 2)
                if (flash % 2) == 0:
                    self.model.model.clearColor()
                else:
                    self.model.model.setColor(0, 0, 1, 1)

        self.model.model.setPos(thisLevel.pixelBoard, self.x + 8, self.y + 8, 0)
        self.model.update()
        
        ## if thisGame.mode == 3:
        ##     return False
        
        
        ## # ghost eyes --
        ## for y in range(4, 8, 1):
        ##     for x in range(3, 7, 1):
        ##         self.anim[ self.animFrame ].set_at( (x, y), (255, 255, 255, 255) )  
        ##         self.anim[ self.animFrame ].set_at( (x+6, y), (255, 255, 255, 255) )
                
        ##         if player.x > self.x and player.y > self.y:
        ##             #player is to lower-right
        ##             pupilSet = (5, 6)
        ##         elif player.x < self.x and player.y > self.y:
        ##             #player is to lower-left
        ##             pupilSet = (3, 6)
        ##         elif player.x > self.x and player.y < self.y:
        ##             #player is to upper-right
        ##             pupilSet = (5, 4)
        ##         elif player.x < self.x and player.y < self.y:
        ##             #player is to upper-left
        ##             pupilSet = (3, 4)
        ##         else:
        ##             pupilSet = (4, 6)
                    
        ## for y in range(pupilSet[1], pupilSet[1] + 2, 1):
        ##     for x in range(pupilSet[0], pupilSet[0] + 2, 1):
        ##         self.anim[ self.animFrame ].set_at( (x, y), (0, 0, 255, 255) )  
        ##         self.anim[ self.animFrame ].set_at( (x+6, y), (0, 0, 255, 255) )    
        ## # -- end ghost eyes
        
        ## if self.stateId == 1:
        ##     # draw regular ghost (this one)
        ##     screen.blit (self.anim[ self.animFrame ], (self.x - thisGame.screenPixelPos[0], self.y - thisGame.screenPixelPos[1]))
        ## elif self.stateId == 2:
        ##     # draw vulnerable ghost
            
        ##     if thisGame.ghostTimer > 100:
        ##         # blue
        ##         screen.blit (ghosts[4].anim[ self.animFrame ], (self.x - thisGame.screenPixelPos[0], self.y - thisGame.screenPixelPos[1]))
        ##     else:
        ##         # blue/white flashing
        ##         tempTimerI = int(thisGame.ghostTimer / 10)
        ##         if tempTimerI == 1 or tempTimerI == 3 or tempTimerI == 5 or tempTimerI == 7 or tempTimerI == 9:
        ##             screen.blit (ghosts[5].anim[ self.animFrame ], (self.x - thisGame.screenPixelPos[0], self.y - thisGame.screenPixelPos[1]))
        ##         else:
        ##             screen.blit (ghosts[4].anim[ self.animFrame ], (self.x - thisGame.screenPixelPos[0], self.y - thisGame.screenPixelPos[1]))
            
        ## elif self.stateId == 3:
        ##     # draw glasses
        ##     screen.blit (tileIDImage[ tileID[ 'glasses' ] ], (self.x - thisGame.screenPixelPos[0], self.y - thisGame.screenPixelPos[1]))
        
        ## if thisGame.mode == 6 or thisGame.mode == 7:
        ##     # don't animate ghost if the level is complete
        ##     return False
        
        ## self.animDelay += 1
        
        ## if self.animDelay == 2:
        ##     self.animFrame += 1 
        
        ##     if self.animFrame == 7:
        ##         # wrap to beginning
        ##         self.animFrame = 1
                
        ##     self.animDelay = 0
            
    def Move (self):
        
        self.x += self.velX
        self.y += self.velY
        self.nearestRow = int(((self.y + 8) / 16))
        self.nearestCol = int(((self.x + 8) / 16))

        if (self.x % 16) == 0 and (self.y % 16) == 0:
            # if the ghost is lined up with the grid again
            # meaning, it's time to go to the next path item
            
            if (self.currentPath):
                self.currentPath = self.currentPath[1:]
                self.FollowNextPathWay()
        
            else:
                self.x = self.nearestCol * 16
                self.y = self.nearestRow * 16
            
                # chase pac-man
                self.currentPath = path.FindPath( (self.nearestRow, self.nearestCol), (player.nearestRow, player.nearestCol) )
                self.FollowNextPathWay()
            
    def FollowNextPathWay (self):
        
        # print "Ghost " + str(self.id) + " rem: " + self.currentPath
        
        # only follow this pathway if there is a possible path found!
        if self.currentPath is not False:
        
            if len(self.currentPath) > 0:
                if self.currentPath[0] == "L":
                    (self.velX, self.velY) = (-self.speed, 0)
                    self.model.setTargetHeading(90)
                elif self.currentPath[0] == "R":
                    (self.velX, self.velY) = (self.speed, 0)
                    self.model.setTargetHeading(270)
                elif self.currentPath[0] == "U":
                    (self.velX, self.velY) = (0, -self.speed)
                    self.model.setTargetHeading(0)
                elif self.currentPath[0] == "D":
                    (self.velX, self.velY) = (0, self.speed)
                    self.model.setTargetHeading(180)
                    
            else:
                # this ghost has reached his destination!!
                
                if self.state != 'Spectacles':
                    # chase pac-man
                    self.currentPath = path.FindPath( (self.nearestRow, self.nearestCol), (player.nearestRow, player.nearestCol) )
                    if len(self.currentPath) > 0:
                        self.FollowNextPathWay()
                
                else:
                    # glasses found way back to ghost box
                    self.request('Normal')
                    self.speed = self.speed / 4
                    
                    # give ghost a path to a random spot (containing a pellet)
                    (randRow, randCol) = (0, 0)

                    while not thisLevel.GetMapTile((randRow, randCol)) == tileID[ 'pellet' ] or (randRow, randCol) == (0, 0):
                        randRow = random.randint(1, thisLevel.lvlHeight - 2)
                        randCol = random.randint(1, thisLevel.lvlWidth - 2)

                    self.currentPath = path.FindPath( (self.nearestRow, self.nearestCol), (randRow, randCol) )
                    self.FollowNextPathWay()

class fruit (FSM):
    def __init__ (self):
        FSM.__init__(self, 'fruit')
        
        # when fruit is not in use, it's in the (-1, -1) position off-screen.
        self.slowTimer = 0
        self.x = -16
        self.y = -16
        self.velX = 0
        self.velY = 0
        self.speed = 1
        self.active = False
        
        self.bouncei = 0
        self.bounceY = 0
        
        self.nearestRow = (-1, -1)
        self.nearestCol = (-1, -1)

        model = loader.loadModel('res/sprite/fruit')
        seq = model.find('**/+SequenceNode')
        frames = seq.getChildren()

        self.model = NodePath('model')
        self.switch = self.model.attachNewNode(SwitchNode('fruit'))
        self.switch.setPos(0, 0, 0.2)
        self.switch.setBillboardPointEye()
        self.switchNode = self.switch.node()
        frames.reparentTo(self.switch)
        self.numFrames = len(frames)

        self.model.setScale(0.05)
        
        self.currentPath = ""

        self.request('Off')
        self.setFruitType(1)

    def setFruitType(self, fruitType):
        assert fruitType >= 0
        self.switchNode.setVisibleChild(fruitType)

    def enterOff(self):
        self.model.detachNode()
        self.active = False

    def enterActive(self):
        self.model.reparentTo(thisLevel.board)
        self.active = True
        
    def Draw (self):
        if self.active:
            self.model.setPos(thisLevel.pixelBoard, self.x + 8, self.y + 8, 0)
            
    def Move (self):
        if not self.active:
            return False
        
        self.bouncei += 1
        if self.bouncei == 1:
            self.bounceY = 2
        elif self.bouncei == 2:
            self.bounceY = 4
        elif self.bouncei == 3:
            self.bounceY = 5
        elif self.bouncei == 4:
            self.bounceY = 5
        elif self.bouncei == 5:
            self.bounceY = 6
        elif self.bouncei == 6:
            self.bounceY = 6
        elif self.bouncei == 9:
            self.bounceY = 6
        elif self.bouncei == 10:
            self.bounceY = 5
        elif self.bouncei == 11:
            self.bounceY = 5
        elif self.bouncei == 12:
            self.bounceY = 4
        elif self.bouncei == 13:
            self.bounceY = 3
        elif self.bouncei == 14:
            self.bounceY = 2
        elif self.bouncei == 15:
            self.bounceY = 1
        elif self.bouncei == 16:
            self.bounceY = 0
            self.bouncei = 0
            snd_fruitbounce.play()
        
        self.slowTimer += 1
        if self.slowTimer == 2:
            self.slowTimer = 0
            
            self.x += self.velX
            self.y += self.velY
            
            self.nearestRow = int(((self.y + 8) / 16))
            self.nearestCol = int(((self.x + 8) / 16))

            if (self.x % 16) == 0 and (self.y % 16) == 0:
                # if the fruit is lined up with the grid again
                # meaning, it's time to go to the next path item
                
                if len(self.currentPath) > 0:
                    self.currentPath = self.currentPath[1:]
                    self.FollowNextPathWay()
            
                else:
                    self.x = self.nearestCol * 16
                    self.y = self.nearestRow * 16
                    
                    self.request('Off')
                    thisGame.fruitTimer = 0
            
    def FollowNextPathWay (self):
        

        # only follow this pathway if there is a possible path found!
        if not self.currentPath == False:
        
            if len(self.currentPath) > 0:
                if self.currentPath[0] == "L":
                    (self.velX, self.velY) = (-self.speed, 0)
                elif self.currentPath[0] == "R":
                    (self.velX, self.velY) = (self.speed, 0)
                elif self.currentPath[0] == "U":
                    (self.velX, self.velY) = (0, -self.speed)
                elif self.currentPath[0] == "D":
                    (self.velX, self.velY) = (0, self.speed)

class pacman ():
    
    def __init__ (self):
        self.x = 0
        self.y = 0
        self.velX = 0
        self.velY = 0
        self.speed = 2
        
        self.nearestRow = 0
        self.nearestCol = 0
        
        self.homeX = 0
        self.homeY = 0

        self.model = threeSixty('batman')
        #self.model.model.setScale(8)
        self.model.model.setScale(0.2)

        self.pelletSndNum = 0
        
    def Move (self):
        
        self.nearestRow = int(((self.y + 8) / 16))
        self.nearestCol = int(((self.x + 8) / 16))

        # make sure the current velocity will not cause a collision before moving
        if not thisLevel.CheckIfHitWall((self.x + self.velX, self.y + self.velY), (self.nearestRow, self.nearestCol)):
            # it's ok to Move
            self.x += self.velX
            self.y += self.velY
            
            # check for collisions with other tiles (pellets, etc)
            thisLevel.CheckIfHitSomething((self.x, self.y), (self.nearestRow, self.nearestCol))
            
            # check for collisions with the ghosts
            for i in range(0, 4, 1):
                if thisLevel.CheckIfHit( (self.x, self.y), (ghosts[i].x, ghosts[i].y), 8):
                    # hit a ghost
                    
                    if ghosts[i].state == 'Normal':
                        # ghost is normal
                        thisGame.request('HitGhost')
                        
                    elif ghosts[i].state == 'Vulnerable':
                        # ghost is vulnerable
                        # give them glasses
                        # make them run
                        thisGame.AddToScore(thisGame.ghostValue, drawAt = (self.x, self.y))
                        thisGame.ghostValue = thisGame.ghostValue * 2
                        snd_eatgh.play()
                        
                        ghosts[i].request('Spectacles')
                        ghosts[i].speed = ghosts[i].speed * 4
                        # and send them to the ghost box
                        ghosts[i].x = ghosts[i].nearestCol * 16
                        ghosts[i].y = ghosts[i].nearestRow * 16
                        ghosts[i].currentPath = path.FindPath( (ghosts[i].nearestRow, ghosts[i].nearestCol), (thisLevel.GetGhostBoxPos()[0]+1, thisLevel.GetGhostBoxPos()[1]) )
                        ghosts[i].FollowNextPathWay()
                        
                        # set game mode to brief pause after eating
                        thisGame.request('WaitAfterEatingGhost')
                        
            # check for collisions with the fruit
            if thisFruit.active:
                if thisLevel.CheckIfHit( (self.x, self.y), (thisFruit.x, thisFruit.y), 8):
                    thisGame.AddToScore(2500, drawAt = (self.x, self.y))
                    thisFruit.request('Off')
                    thisGame.fruitTimer = 0
                    thisGame.fruitScoreTimer = 120
                    snd_eatfruit.play()
        
        else:
            # we're going to hit a wall -- stop moving
            self.velX = 0
            self.velY = 0
            
        ## # deal with power-pellet ghost timer
        ## if thisGame.ghostTimer > 0:
        ##     thisGame.ghostTimer -= 1
            
        ##     if thisGame.ghostTimer == 0:
        ##         for i in range(0, 4, 1):
        ##             if ghosts[i].stateId == 2:
        ##                 ghosts[i].stateId = 1
        ##         self.ghostValue = 0
                
        # deal with fruit timer
        thisGame.fruitTimer += 1
        if thisGame.fruitTimer == 500:
            pathwayPair = thisLevel.GetPathwayPairPos()
            
            if not pathwayPair == False:
            
                pathwayEntrance = pathwayPair[0]
                pathwayExit = pathwayPair[1]
                
                thisFruit.request('Active')
                
                thisFruit.nearestRow = pathwayEntrance[0]
                thisFruit.nearestCol = pathwayEntrance[1]
                
                thisFruit.x = thisFruit.nearestCol * 16
                thisFruit.y = thisFruit.nearestRow * 16
                
                thisFruit.currentPath = path.FindPath( (thisFruit.nearestRow, thisFruit.nearestCol), pathwayExit )
                thisFruit.FollowNextPathWay()
            
        if thisGame.fruitScoreTimer > 0:
            thisGame.fruitScoreTimer -= 1
            
        
    def Draw (self):
        if thisGame.state == 'GameOver':
            self.model.model.detachNode()
            return

        self.model.model.reparentTo(thisLevel.board)
        self.model.model.setPos(thisLevel.pixelBoard, self.x + 8, self.y + 8, 0)
        #self.model.setScale(thisLevel.board, 8)
        self.model.update()
            
class level ():
    
    def __init__ (self):
        self.lvlWidth = 0
        self.lvlHeight = 0
        self.edgeLightColor = (255, 255, 0, 255)
        self.edgeShadowColor = (255, 150, 0, 255)
        self.fillColor = (0, 255, 255, 255)

        # For flashing at the end of the level.
        self.altEdgeLightColor = (255, 255, 255, 255)
        self.altEdgeShadowColor = (255, 255, 255, 255)
        self.altFillColor = (0, 0, 0, 255)

        self.pelletColor = (255, 255, 255, 255)

        self.map = {}
        self.mapCard = None
        
        self.pellets = 0
        self.powerPelletBlinkTimer = 0
        self.pelletModels = {}

        self.board = thisGame.board
        self.board.setP(30)
        self.board.setPos(0, 20, 0)
        self.board.setScale(10)

        self.pixelBoard = self.board.attachNewNode('pixelBoard')

    def setPellet(self, (row, col), power):
        x = (col + 0.5) / float(self.lvlWidth) - 0.5
        y = 0.5 - (row + 0.5) / float(self.lvlHeight)
        if power:
            model = loader.loadModel('res/sprite/power_pellet')
        else:
            model = loader.loadModel('res/sprite/pellet')
        model.reparentTo(self.board)
        model.setPos(x, y, 0)
        self.pelletModels[(row, col)] = model

    def clearPellet(self, (row, col)):
        if (row, col) in self.pelletModels:
            self.pelletModels[(row, col)].detachNode()
        
    def SetMapTile (self, (row, col), newValue):
        self.map[ (row * self.lvlWidth) + col ] = newValue
        
    def GetMapTile (self, (row, col)):
        if row >= 0 and row < self.lvlHeight and col >= 0 and col < self.lvlWidth:
            return self.map[ (row * self.lvlWidth) + col ]
        else:
            return 0
    
    def IsWall (self, (row, col)):
    
        if row > thisLevel.lvlHeight - 1 or row < 0:
            return True
        
        if col > thisLevel.lvlWidth - 1 or col < 0:
            return True
    
        # check the offending tile ID
        result = thisLevel.GetMapTile((row, col))

        # if the tile was a wall
        if result >= 100 and result <= 199:
            return True
        else:
            return False
    
                    
    def CheckIfHitWall (self, (possiblePlayerX, possiblePlayerY), (row, col)):
    
        numCollisions = 0
        
        # check each of the 9 surrounding tiles for a collision
        for iRow in range(row - 1, row + 2, 1):
            for iCol in range(col - 1, col + 2, 1):
            
                if  (possiblePlayerX - (iCol * 16) < 16) and (possiblePlayerX - (iCol * 16) > -16) and (possiblePlayerY - (iRow * 16) < 16) and (possiblePlayerY - (iRow * 16) > -16):
                    
                    if self.IsWall((iRow, iCol)):
                        numCollisions += 1
                        
        if numCollisions > 0:
            return True
        else:
            return False
        
        
    def CheckIfHit (self, (playerX, playerY), (x, y), cushion):
    
        if (playerX - x < cushion) and (playerX - x > -cushion) and (playerY - y < cushion) and (playerY - y > -cushion):
            return True
        else:
            return False


    def CheckIfHitSomething (self, (playerX, playerY), (row, col)):
    
        for iRow in range(row - 1, row + 2, 1):
            for iCol in range(col - 1, col + 2, 1):
            
                if  (playerX - (iCol * 16) < 16) and (playerX - (iCol * 16) > -16) and (playerY - (iRow * 16) < 16) and (playerY - (iRow * 16) > -16):
                    # check the offending tile ID
                    result = thisLevel.GetMapTile((iRow, iCol))
        
                    if result == tileID[ 'pellet' ]:
                        # got a pellet
                        thisLevel.SetMapTile((iRow, iCol), 0)
                        thisLevel.clearPellet((iRow, iCol))
                        snd_pellet[player.pelletSndNum].play()
                        player.pelletSndNum = 1 - player.pelletSndNum
                        
                        thisLevel.pellets -= 1
                        
                        thisGame.AddToScore(10)
                        
                        if thisLevel.pellets == 0:
                            # no more pellets left!
                            # WON THE LEVEL
                            thisGame.request('WaitAfterFinishing')
                            
                        
                    elif result == tileID[ 'pellet-power' ]:
                        # got a power pellet
                        thisLevel.SetMapTile((iRow, iCol), 0)
                        thisLevel.clearPellet((iRow, iCol))
                        snd_powerpellet.play()
                        
                        thisGame.AddToScore(100)
                        thisGame.ghostValue = 200
                        
                        thisGame.ghostTimer = 360
                        for i in range(0, 4, 1):
                            ghosts[i].request('Vulnerable')
                        
                    elif result == tileID[ 'door-h' ]:
                        # ran into a horizontal door
                        for i in range(0, thisLevel.lvlWidth, 1):
                            if not i == iCol:
                                if thisLevel.GetMapTile((iRow, i)) == tileID[ 'door-h' ]:
                                    player.x = i * 16
                                    
                                    if player.velX > 0:
                                        player.x += 16
                                    else:
                                        player.x -= 16
                                        
                    elif result == tileID[ 'door-v' ]:
                        # ran into a vertical door
                        for i in range(0, thisLevel.lvlHeight, 1):
                            if not i == iRow:
                                if thisLevel.GetMapTile((i, iCol)) == tileID[ 'door-v' ]:
                                    player.y = i * 16
                                    
                                    if player.velY > 0:
                                        player.y += 16
                                    else:
                                        player.y -= 16
                                        
    def GetGhostBoxPos (self):
        
        for row in range(0, self.lvlHeight, 1):
            for col in range(0, self.lvlWidth, 1):
                if self.GetMapTile((row, col)) == tileID[ 'ghost-door' ]:
                    return (row, col)
                
        return False
    
    def GetPathwayPairPos (self):
        
        doorArray = []
        
        for row in range(0, self.lvlHeight, 1):
            for col in range(0, self.lvlWidth, 1):
                if self.GetMapTile((row, col)) == tileID[ 'door-h' ]:
                    # found a horizontal door
                    doorArray.append( (row, col) )
                elif self.GetMapTile((row, col)) == tileID[ 'door-v' ]:
                    # found a vertical door
                    doorArray.append( (row, col) )
        
        if len(doorArray) == 0:
            return False
        
        chosenDoor = random.randint(0, len(doorArray) - 1)
        
        if self.GetMapTile( doorArray[chosenDoor] ) == tileID[ 'door-h' ]:
            # horizontal door was chosen
            # look for the opposite one
            for i in range(0, thisLevel.lvlWidth, 1):
                if not i == doorArray[chosenDoor][1]:
                    if thisLevel.GetMapTile((doorArray[chosenDoor][0], i)) == tileID[ 'door-h' ]:
                        return doorArray[chosenDoor], (doorArray[chosenDoor][0], i)
        else:
            # vertical door was chosen
            # look for the opposite one
            for i in range(0, thisLevel.lvlHeight, 1):
                if not i == doorArray[chosenDoor][0]:
                    if thisLevel.GetMapTile((i, doorArray[chosenDoor][1])) == tileID[ 'door-v' ]:
                        return doorArray[chosenDoor], (i, doorArray[chosenDoor][1])
                    
        return False
        
    def PrintMap (self):
        
        for row in range(0, self.lvlHeight, 1):
            outputLine = ""
            for col in range(0, self.lvlWidth, 1):
            
                outputLine += str( self.GetMapTile((row, col)) ) + ", "
                
            print outputLine
            
    def DrawMap (self):
        #self.PaintMap()
        pass

    def PaintMap(self):
        if self.mapCard:
            self.mapCard.removeNode()

        pixelWidth = thisLevel.lvlWidth * 16
        pixelHeight = thisLevel.lvlHeight * 16

        p0 = PNMImage(pixelWidth, pixelHeight, 4)
        p1 = PNMImage(pixelWidth, pixelHeight, 4)
        for row in range(thisLevel.lvlHeight):
            for col in range(thisLevel.lvlWidth):
                useTile = self.GetMapTile((row, col))
                if not useTile == 0 and not useTile == tileID['door-h'] and not useTile == tileID['door-v']:
                    # if this isn't a blank tile
                    if useTile == tileID['pellet']:
                        self.setPellet((row, col), False)

                    elif useTile == tileID['pellet-power']:
                        self.setPellet((row, col), True)

                    elif useTile == tileID['showlogo']:
                        pass
                    
                    elif useTile == tileID['hiscores']:
                        pass
                    
                    else:
                        p0.copySubImage(tileIDImage[ useTile, 0 ],
                                        col * 16, row * 16)
                        p1.copySubImage(tileIDImage[ useTile, 1 ],
                                        col * 16, row * 16)

        p2Width = Texture.upToPower2(pixelWidth)
        p2Height = Texture.upToPower2(pixelHeight)
        p0w = PNMImage(p2Width, p2Height, 4)
        p1w = PNMImage(p2Width, p2Height, 4)
        p0w.quickFilterFrom(p0)
        p1w.quickFilterFrom(p1)

        self.mapTexture0 = Texture('map0')
        self.mapTexture0.load(p0w)
        self.mapTexture0.setMinfilter(Texture.FTLinearMipmapLinear)
        self.mapTexture0.setWrapU(Texture.WMClamp)
        self.mapTexture0.setWrapV(Texture.WMClamp)
        self.mapTexture1 = Texture('map1')
        self.mapTexture1.load(p1w)
        self.mapTexture1.setMinfilter(Texture.FTLinearMipmapLinear)
        self.mapTexture1.setWrapU(Texture.WMClamp)
        self.mapTexture1.setWrapV(Texture.WMClamp)
        self.mapCard = loader.loadModel('res/tiles/card')
        self.mapCard.setTwoSided(True)
        self.mapCard.setTexture(self.mapTexture0, 1)
        self.mapCard.reparentTo(self.board)
        self.mapCard.setPos(-0.5, -0.5, 0)
        #self.mapCard.setTransparency(0, 1)

        self.pixelBoard.setScale(1.0 / pixelWidth, -1.0 / pixelHeight, -1.0 / pixelHeight)
        self.pixelBoard.setPos(-0.5, 0.5, 0)
        
    def LoadLevel (self, levelNum):
        
        self.map = {}
        
        self.pellets = 0
        for model in self.pelletModels.values():
            model.detachNode()
        self.pelletModels = {}

        levelNum = (levelNum % 13)
        f = open(SCRIPT_PATH + "/res/levels/"+str(levelNum) + ".txt", 'r')
            
        # ANDY -- edit this
        #fileOutput = f.read()
        #str_splitByLine = fileOutput.split('\n')
        lineNum=-1
        rowNum = 0
        useLine = False
        isReadingLevelData = False
          
        for line in f:

          lineNum += 1
        
            # print " ------- Level Line " + str(lineNum) + " -------- "
          while len(line)>0 and (line[-1]=="\n" or line[-1]=="\r"): line=line[:-1]
          while len(line)>0 and (line[0]=="\n" or line[0]=="\r"): line=line[1:]
          str_splitBySpace = line.split(' ')
            
            
          j = str_splitBySpace[0]
                
          if (j == "'" or j == ""):
                # comment / whitespace line
                # print " ignoring comment line.. "
                useLine = False
          elif j == "#":
                # special divider / attribute line
                useLine = False
                
                firstWord = str_splitBySpace[1]
                
                if firstWord == "lvlwidth":
                    self.lvlWidth = int( str_splitBySpace[2] )
                    # print "Width is " + str( self.lvlWidth )
                    
                elif firstWord == "lvlheight":
                    self.lvlHeight = int( str_splitBySpace[2] )
                    # print "Height is " + str( self.lvlHeight )
                    
                elif firstWord == "edgecolor":
                    # edge color keyword for backwards compatibility (single edge color) mazes
                    red = int( str_splitBySpace[2] )
                    green = int( str_splitBySpace[3] )
                    blue = int( str_splitBySpace[4] )
                    self.edgeLightColor = (red, green, blue, 255)
                    self.edgeShadowColor = (red, green, blue, 255)
                    
                elif firstWord == "edgelightcolor":
                    red = int( str_splitBySpace[2] )
                    green = int( str_splitBySpace[3] )
                    blue = int( str_splitBySpace[4] )
                    self.edgeLightColor = (red, green, blue, 255)
                    
                elif firstWord == "edgeshadowcolor":
                    red = int( str_splitBySpace[2] )
                    green = int( str_splitBySpace[3] )
                    blue = int( str_splitBySpace[4] )
                    self.edgeShadowColor = (red, green, blue, 255)
                
                elif firstWord == "fillcolor":
                    red = int( str_splitBySpace[2] )
                    green = int( str_splitBySpace[3] )
                    blue = int( str_splitBySpace[4] )
                    self.fillColor = (red, green, blue, 255)
                    
                elif firstWord == "pelletcolor":
                    red = int( str_splitBySpace[2] )
                    green = int( str_splitBySpace[3] )
                    blue = int( str_splitBySpace[4] )
                    self.pelletColor = (red, green, blue, 255)
                    
                elif firstWord == "fruittype":
                    thisFruit.setFruitType( int(str_splitBySpace[2]) )
                    
                elif firstWord == "startleveldata":
                    isReadingLevelData = True
                        # print "Level data has begun"
                    rowNum = 0
                    
                elif firstWord == "endleveldata":
                    isReadingLevelData = False
                    # print "Level data has ended"
                    
          else:
                useLine = True
                
                
            # this is a map data line   
          if useLine == True:
                
                if isReadingLevelData == True:
                        
                    # print str( len(str_splitBySpace) ) + " tiles in this column"
                    
                    for k in range(0, self.lvlWidth, 1):
                        self.SetMapTile((rowNum, k), int(str_splitBySpace[k]) )
                        
                        thisID = int(str_splitBySpace[k])
                        if thisID == 4: 
                            # starting position for pac-man
                            
                            player.homeX = k * 16
                            player.homeY = rowNum * 16
                            self.SetMapTile((rowNum, k), 0 )
                            
                        elif thisID >= 10 and thisID <= 13:
                            # one of the ghosts
                            
                            ghosts[thisID - 10].homeX = k * 16
                            ghosts[thisID - 10].homeY = rowNum * 16
                            self.SetMapTile((rowNum, k), 0 )
                        
                        elif thisID == 2:
                            # pellet
                            
                            self.pellets += 1
                            
                    rowNum += 1
                    
                
        # reload all tiles and set appropriate colors
        GetCrossRef()

        # load map into the pathfinder object
        path.ResizeMap( (self.lvlHeight, self.lvlWidth) )
        
        for row in range(0, path.size[0], 1):
            for col in range(0, path.size[1], 1):
                if self.IsWall( (row, col) ):
                    path.SetType( (row, col), 1 )
                else:
                    path.SetType( (row, col), 0 )
        
        # do all the level-starting stuff
        self.Restart()
        self.PaintMap()

    def Restart (self):
        
        for i in range(0, 4, 1):
            # move ghosts back to home

            ghosts[i].x = ghosts[i].homeX
            ghosts[i].y = ghosts[i].homeY
            ghosts[i].velX = 0
            ghosts[i].velY = 0
            ghosts[i].request('Normal')
            ghosts[i].speed = 1
            ghosts[i].Move()
            
            # give each ghost a path to a random spot (containing a pellet)
            (randRow, randCol) = (0, 0)

            while not self.GetMapTile((randRow, randCol)) == tileID[ 'pellet' ] or (randRow, randCol) == (0, 0):
                randRow = random.randint(1, self.lvlHeight - 2)
                randCol = random.randint(1, self.lvlWidth - 2)
            
            # print "Ghost " + str(i) + " headed towards " + str((randRow, randCol))
            ghosts[i].currentPath = path.FindPath( (ghosts[i].nearestRow, ghosts[i].nearestCol), (randRow, randCol) )
            ghosts[i].FollowNextPathWay()
            
        thisFruit.request('Off')
            
        thisGame.fruitTimer = 0

        player.x = player.homeX
        player.y = player.homeY
        player.velX = 0
        player.velY = 0
        
        player.animFrame = 3



# Set of keys currently held.
keys = set()

# Set up handlers to manage the keys set, above.
def buttonDownEvent(keyname):
    keys.add(keyname)
def buttonUpEvent(keyname):
    if keyname in keys:
        keys.remove(keyname)
bt = base.buttonThrowers[0].node()
bt.setModifierButtons(ModifierButtons())
bt.setButtonDownEvent('buttonDown')
base.accept('buttonDown', buttonDownEvent)
bt.setButtonUpEvent('buttonUp')
base.accept('buttonUp', buttonUpEvent)

def CheckInputs(): 
    if 'arrow_right' in keys:
        if not thisLevel.CheckIfHitWall((player.x + player.speed, player.y), (player.nearestRow, player.nearestCol)): 
            player.velX = player.speed
            player.velY = 0
            player.model.setTargetHeading(270)

    if 'arrow_left' in keys:
        if not thisLevel.CheckIfHitWall((player.x - player.speed, player.y), (player.nearestRow, player.nearestCol)): 
            player.velX = -player.speed
            player.velY = 0
            player.model.setTargetHeading(90)

    if 'arrow_down' in keys:
        if not thisLevel.CheckIfHitWall((player.x, player.y + player.speed), (player.nearestRow, player.nearestCol)): 
            player.velX = 0
            player.velY = player.speed
            player.model.setTargetHeading(180)

    if 'arrow_up' in keys:
        if not thisLevel.CheckIfHitWall((player.x, player.y - player.speed), (player.nearestRow, player.nearestCol)):
            player.velX = 0
            player.velY = -player.speed
            player.model.setTargetHeading(0)
    
#      _____________________________________________
# ___/  function: Get ID-Tilename Cross References  \______________________________________ 
    
def GetCrossRef ():

    f = open(SCRIPT_PATH + "/res/crossref.txt", 'r')
    # ANDY -- edit
    #fileOutput = f.read()
    #str_splitByLine = fileOutput.split('\n')

    lineNum = 0
    useLine = False

    for i in f.readlines():
        # print " ========= Line " + str(lineNum) + " ============ "
        while len(i)>0 and (i[-1]=='\n' or i[-1]=='\r'): i=i[:-1]
        while len(i)>0 and (i[0]=='\n' or i[0]=='\r'): i=i[1:]
        str_splitBySpace = i.split(' ')
        
        j = str_splitBySpace[0]
            
        if (j == "'" or j == "" or j == "#"):
            # comment / whitespace line
            # print " ignoring comment line.. "
            useLine = False
        else:
            # print str(wordNum) + ". " + j
            useLine = True
        
        if useLine == True:
            tileIDName[ int(str_splitBySpace[0]) ] = str_splitBySpace[1]
            tileID[ str_splitBySpace[1] ] = int(str_splitBySpace[0])
            
            thisID = int(str_splitBySpace[0])
            if not thisID in NO_GIF_TILES:
                tileIDImage[ thisID, 0 ] = PNMImage(SCRIPT_PATH + "/res/tiles/"+str_splitBySpace[1] + ".png")
            else:
                tileIDImage[ thisID, 0 ] = PNMImage(16, 16, 4)

            tileIDImage[ thisID, 1 ] = PNMImage(tileIDImage[ thisID, 0 ])
            
            # change colors in tileIDImage to match maze colors
            wallEdgeColor = PNMImage.PixelSpec(255, 206, 255, 255)
            wallFillColor = PNMImage.PixelSpec(132, 0, 132, 255)
            edgeShadowColor = PNMImage.PixelSpec(255, 0, 255, 255)
            pelletColor = PNMImage.PixelSpec(128, 0, 128, 255)
            for y in range(0, 16, 1):
                for x in range(0, 16, 1):
                    
                    if tileIDImage[ thisID, 0 ].getPixel(x, y) == wallEdgeColor:
                        # wall edge
                        tileIDImage[ thisID, 0 ].setPixel(x, y, thisLevel.edgeLightColor)
                        tileIDImage[ thisID, 1 ].setPixel(x, y, thisLevel.altEdgeLightColor)
                        
                    elif tileIDImage[ thisID, 0 ].getPixel(x, y) == wallFillColor:
                        # wall fill
                        tileIDImage[ thisID, 0 ].setPixel(x, y, thisLevel.fillColor ) 
                        tileIDImage[ thisID, 1 ].setPixel(x, y, thisLevel.altFillColor ) 
                        
                    elif tileIDImage[ thisID, 0 ].getPixel(x, y) == edgeShadowColor:
                        # edge shadow
                        tileIDImage[ thisID, 0 ].setPixel(x, y, thisLevel.edgeShadowColor )   
                        tileIDImage[ thisID, 1 ].setPixel(x, y, thisLevel.altEdgeShadowColor )   
                        
                    elif tileIDImage[ thisID, 0 ].getPixel(x, y) == pelletColor:
                        # pellet color
                        tileIDImage[ thisID, 0 ].setPixel(x, y, thisLevel.pelletColor )   
                        tileIDImage[ thisID, 1 ].setPixel(x, y, thisLevel.pelletColor )   
                
            # print str_splitBySpace[0] + " is married to " + str_splitBySpace[1]
        lineNum += 1


#      __________________
# ___/  main code block  \_____________________________________________________


base.disableMouse()
base.setBackgroundColor((0, 0, 0, 1))

# create the pacman
player = pacman()

# create a path_finder object
path = path_finder()

# create ghost objects
ghosts = {}
for i in range(0, 6, 1):
    # remember, ghost[4] is the blue, vulnerable ghost
    ghosts[i] = ghost(i)
    
# create piece of fruit
thisFruit = fruit()

tileIDName = {} # gives tile name (when the ID# is known)
tileID = {} # gives tile ID (when the name is known)
tileIDImage = {} # gives tile image (when the ID# is known)

# create game and level objects and load first level
thisGame = game()
thisLevel = level()
thisLevel.LoadLevel( thisGame.GetLevelNum() )
js = None

def Move():
    player.Move()
    for i in range(0, 4, 1):
        ghosts[i].Move()
    thisFruit.Move()

def gameTask(task):

    if thisGame.state == 'Normal':
        # normal gameplay mode
        CheckInputs()

        # Call Move() exactly 60 times per second.
        thisFrame = int(globalClock.getFrameTime() * 60.0)
        elapsed = thisFrame - thisGame.lastMoveFrame
        thisGame.lastMoveFrame = thisFrame
        for i in range(elapsed):
            Move()

    thisLevel.DrawMap()

    for i in range(0, 4, 1):
        ghosts[i].Draw()
    thisFruit.Draw()
    player.Draw()
    
    thisGame.DrawScore()
    
    return task.cont

taskMgr.add(gameTask, 'gameTask')

run()

